/* bf.c
******************************************************************************

   Yet another Brainfuck interpreter

   Author: Stephan Beyer
   
   Copyright (C) GPL, 2003, 2004 Stephan Beyer - s-beyer@gmx.net
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h> /* malloc, atoi */
#include <string.h> /* strlen */
#include <signal.h> /* signal */

#include "bf.h"
#include "errors.h"
#include "tty.h"

#ifndef ARRAYSTACK
# include "stack.h"
#endif

#define	ShowCode(c)	if (opt.showinput) fprintf(stderr, "%c", c);

enum operations
{
	opPLUS = 0x0,	/* + = 000b */
	opMINUS,	/* - = 001b */
	opGREATER,	/* > = 010b */
	opLESS,		/* < = 011b */
	opBEGIN,	/* [ = 100b */
	opEND,		/* ] = 101b */
	opOUTPUT,	/* . = 110b */
	opINPUT		/* , = 111b */
};

/* global variables: */
char *array; /* the Brainfuck array */
char *p; /* array pointer */
char *program;
/* We put the brainfuck code 'compressed' into memory before interpreting
 * it. for that we need our global variable program.
 * One program char (containing of 8 bits) can be splitted into:
 *   - amount of Brainfuck operator - 1, 5 bits (max number: 32)
 *   - Brainfuck operator, 3 bits (see operations enum)
 *   - opBEGINs and opENDs mustn't be compressed (amount bits must be 00000b)
 */
int psize; /* program size (because there's no termination character) */
#ifdef ARRAYSTACK
int stacksize; /* fixed size of stack (brackets) */
#endif

/* SIGINT, SIGTERM, SIGQUIT handler */
void ForceExit(int signum)
{
	ttyRestore(); /* if signal during "," */
	switch(signum)
	{
		case SIGINT:
			puts("Interrupted...");
			break;
		case SIGTERM:
			puts("Terminated...");
			break;
		case SIGQUIT:
			puts("Quit...");
			break;
	}
	exit(errSIGNAL);
}

/* put <j> <c>-characters at position <i> into <program>
 * 1 <= j <= 32 ; return value is the new position */
int PutIntoProgram(char c, char j, int i)
{
	*(program+i) = ((j-1) << 3);
	switch(c)
	{
		case '+':
			*(program+i) |= opPLUS;
			break;
		case '-':
			*(program+i) |= opMINUS;
			break;
		case '<':
			*(program+i) |= opLESS;
			break;
		case '>':
			*(program+i) |= opGREATER;
			break;
		case '[':
			for(; j > 0; j--)
			{
				*(program+i) = opBEGIN;
				i++;
			}
			i--;
			break;
		case ']':
			for(; j > 0; j--)
			{
				*(program+i) = opEND;
				i++;
			}
			i--;
			break;
		case '.':
			*(program+i) |= opOUTPUT;
			break;
		case ',':
			*(program+i) |= opINPUT;
			break;
	}
	i++;
	return i;
}

/* read input file and write it into <program> */
/* also calculate stack size */
void ReadProgram(void)
{
	int stage; /* read stage (to avoid double code) */
	FILE *fp;
	int i;
	
	int brackets = 0;
#ifdef ARRAYSTACK
	stacksize = 0;
#endif

	/* file opening banana */
	fp = fopen(opt.filename, "r");
	DoAndErr(!fp, fopen, errFILE); 

	for(stage = 0; stage <= 1; stage++)
	{
		int c; /* fgetc()ed char */
		char oldc = 0;
		char j = 0;
		i = 0;

		rewind(fp);

		/* stage 0: determine program size -> alloc mem
		 * stage 1: write program into mem */
		while ((c = fgetc(fp)) != EOF)
		{
			if ((c == '+') || (c == '-')
			 || (c == '<') || (c == '>')
			 || (c == '.') || (c == ',')
			 || (c == '[') || (c == ']'))
			{
				if (stage && !oldc) /* stage 1: init oldc */
					oldc = c;
				if ((oldc != c) || (j >= 32))
				{
					if (!stage) /* stage 0 */
					{
						/* [] must not be compressed */
						if ((oldc == '[')
						 || (oldc == ']'))
							i += j - 1;
						i++; /* increment size */
					} else /* stage 1 */
					{
						i = PutIntoProgram(oldc, j, i);
						if (oldc == '[')
						{
							brackets += j;
#ifdef ARRAYSTACK
							if (brackets > stacksize)
								stacksize = brackets;
#endif
						} else
						if (oldc == ']')
							brackets -= j;
					}
					j = 1; /* this char is there one time */
				}
				else
					j++; /* increment amount of one char */
				oldc = c;
			} 
		}
		
		if (!stage) /* stage 0 */
		{
			/* alloc memory for program */
			psize = i; /* save global psize */
			program = malloc(psize);
			DoAndErr(!program, malloc, errMEMORY);
		} else { /* stage 1 */
			PutIntoProgram(oldc, j, i);
			if (oldc == ']')
				brackets--;
		}
	}
	if (brackets)
		ErrorMsg(errBRACKET, "Unbalanced brackets.");
	fclose(fp);
}

char GetInput(void)
{
	char c;
	if (!opt.inputmode)
		c = getchar();
	else
		c = ttyGetInput();
	if (opt.null && (c == '\n')) /* translate \n to \0? */
		c = 0;
	return c;
}

void Interprete(void)
{
	int i = 0; /* the char in the program */
	int do_nothing = 0;

#ifdef ARRAYSTACK
	int stack[stacksize];
	int s = 0; /* stack index */

	/* first: reset stack */
	for (i=0; i<stacksize; i++)
		stack[i]=0;
	i = 0;
#else
	Stack *s = NULL; /* reset stack */
#endif

	/* now interprete */
	while (i < psize)
	{
		int k = 0;
		char c = *(program + i);
		int num = (c >> 3) + 1; /* number of times to repeat */
		c &= 0x07; /* the operation code (3 bit) */
		i++;

		if (do_nothing)
		{
			if (c == opEND)
				do_nothing--;
			else
			if (c == opBEGIN)
				do_nothing++;
			continue;
		}

		switch(c)
		{
			case opGREATER:
				if (p+num > array+opt.cells)
					ErrorMsg(errOUTOFRANGE, "Out of range! You wanted to '>' beyond the last cell. See -c option.");
				p += num;
				ShowCode('>');
				break;
			case opLESS:
				if (p-num < array)
					ErrorMsg(errOUTOFRANGE, "Out of range! You wanted to '<' below the first cell.\nTo solve, add some '>'s at the beginning, for example.");
				p -= num;
				ShowCode('<');
				break;
			case opPLUS:
				if ((!opt.wraparound) && (*p+num > 255))
					ErrorMsg(errWRAPAROUND, "Out of range! You wanted to '+' a 0xFF byte. See -w option.");
				*p += num;
				ShowCode('+');
				break;
			case opMINUS:
				if ((!opt.wraparound) && (*p-num < 0))
					ErrorMsg(errWRAPAROUND, "Out of range! You wanted to '-' a 0x00 byte. See -w option.");
				*p -= num;
				ShowCode('-');
				break;
			case opOUTPUT:
				for (k = 0; k < num; k++)
				{
					ShowCode('.');
					putchar(*p);
				}
				fflush(stdout);
				break;
			case opINPUT:
				for (k = 0; k < num; k++)
				{
					ShowCode(',');
					*p = GetInput();
				}
				break;
			case opBEGIN:
				ShowCode('[');
				if (!*p)
				{
					do_nothing++;
					Debug(skip loop)
					break;
				}
				Debug(begin loop);
				/* push */
#ifdef ARRAYSTACK
				stack[s] = i;
				s++;
#else
				stackPush(&s, i);
#endif
				break;
			case opEND:
				ShowCode(']');
				if (*p) /* repeat */
				{
					Debug(loop);
#ifdef ARRAYSTACK
					i = stack[s-1];
#else
					i = s->i;
#endif
				} else { /* end loops */
					/* pop */
#ifdef ARRAYSTACK
					s--;
					stack[s] = 0;
#else
					stackPop(&s);
#endif
					Debug(loop ended);
				}
				break;
		}
	}
}

void ResetArray(void)
{
	int i;
	array = malloc(opt.cells+1);
	DoAndErr(!array, malloc, errMEMORY);
	p = array;

	for (i = 0; i < opt.cells; i++)
		*(p++) = 0;

	p = array;
}

/* display usage information */
void Usage(char *bin)
{
	printf(	"Yet another Brainfuck interpreter       version %s\n", VERSION);
	puts(	"(C) 2003, 2004, Stephan Beyer, GPL, s-beyer@gmx.net\n\n"
		"Usage information: ");
	printf(	"\t%s [-h] [options] inputfile\n\n", bin);
	puts(	"Available options:");
	printf(	"\t-c<num>   specify number of cells [%d]\n", opt.cells);
	printf(	"\t-d        show some debugging information (stderr)\n"
		"\t-i        show used code input (stderr)\n"
		"\t-n        translate input: 10 (\\n) to 0\n"
		"\t-w        disallow decrementing 0 and incrementing 255\n"
		"\t-,<mode>  set input mode: 0-4 [%d]\n", opt.inputmode);
	puts(	"\nSee the bf(1) manpage for more information.\n"
		"Have fun!");
}

void HandleOptions(int c, char **v)
/* too lazy to use getopt this time ;) */
{
	int i;
	/* setting defaults */
	opt.showinput = 0;
	opt.cells = CELLS;
	opt.inputmode = 0;
	opt.debug = 0;
	opt.null = 0;
	opt.wraparound = 1; /* allow by default */
	
	if (c <= 1) /* no filename given */
	{
		Usage(*v);
		ErrorMsg(errOPT, "No input file given.");
	}
	if (!strcmp(*(v+1), "-h")) /* is first argument '-h'? */
	{
		Usage(*v);
		exit(0);
	}
	
	for(i = 1; i < c-1; i++) /* options? */
	{
		if ((**(v+i) == '-')
		 && (strlen(*(v+i)) >= 2))
		{
			switch(*(*(v+i)+1))
			{
				case 'c':
					opt.cells = atoi(*(v+i)+2);
					break;
				case 'd':
					opt.debug = 1;
					break;
				case 'i':
					opt.showinput = 1;
					break;
				case 'n':
					opt.null = 1;
					break;
				case 'w':
					opt.wraparound = 0;
					break;
				case ',':
					opt.inputmode = atoi(*(v+i)+2);
					break;
				default:
					Usage(*v);
					printf("Unable to handle option %s - ",
						*(v+i));
					ErrorMsg(errOPT, "Unknown option.");
			}
		} else {
			Usage(*v);
			printf("Invalid argument %s - ", *(v+i));
			ErrorMsg(errOPT, "Invalid argument(s).");
		}
	}
	opt.filename = *(v+i); /* last option MUST be filename */
}

int main(int argc, char **argv)
{
	HandleOptions(argc, argv);
	ReadProgram();
	ResetArray();
	ttyInit();

	/* catch some signals */
	/* TODO use sigaction? */
	signal(SIGINT, ForceExit);
	signal(SIGTERM, ForceExit);
	signal(SIGQUIT, ForceExit);

	/* reading and interpreting */
	Interprete();

	/* cleanup */
	free(array);
	free(program);
	return errNO;
}

/* TODO list
 * - add missing features needed for Brainfuck contests, if there are some
 * - make it DOS-compatible?
 */
